using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;
using MeshVR;
using MacGruber;
using UnityEngine;
using SimpleJSON;
using MVR.FileManagementSecure;


namespace CheesyFX {
	public partial class CheesyFX_WorkOut : MVRScript {
        TextureManager textureManager;
        public List<Rigidbody> forceReceivers = new List<Rigidbody>();
        float movementIntensity;
        public override void Init(){
            FileManagerSecure.CreateDirectory(BASE_DIRECTORY);
            loadURL = new JSONStorableUrl("loadURL", "", UILoadJSON, "jpg", true);
            loadURL.hideExtension = true;
            loadURL.setCallbackFunction -= UILoadJSON;
            loadURL.allowFullComputerBrowse = false;
            loadURL.allowBrowseAboveSuggestedPath = true;
            loadURL.SetFilePath(BASE_DIRECTORY+"/");
            loadURL.setCallbackFunction += UILoadJSON;
            foreach(ForceReceiver fr in containingAtom.forceReceivers){
				Rigidbody rb = fr.GetComponent<Rigidbody>();
				forceReceivers.Add(rb);
			}
            textureManager = new TextureManager(containingAtom);
            textureManager.drivers.Add(new FloatDriver(GetMovementIntensity, textureManager.glossJSON));
            Utils.OnInitUI(CreateUIElement);
			CreateUI();
        }

        public float GetMovementIntensity(){
			float intensity = 0f;
			foreach(Rigidbody rb in forceReceivers){
				intensity += rb.mass * (rb.GetAcceleration() + rb.GetAngularAcceleration()).magnitude;
			}
			movementIntensity = intensity / 1000f;
			return movementIntensity;
		}

        

        public void FixedUpdate(){
            textureManager.drivers.ForEach(x => x.Update());
        }
    }

    public class TextureManager
	{
		public Atom atom;
		public JSONStorable skin;
		public JSONStorableFloat glossJSON;
		public List<FloatDriver> drivers = new List<FloatDriver>();

		public TextureManager(Atom atom){
            this.atom = atom;
			skin = GetStorable("skin");
			glossJSON = skin.GetFloatJSONParam("Gloss");
		}

        public JSONStorable GetStorable(string name){
			string storableId = atom.GetStorableIDs().FirstOrDefault(x => x.Contains(name));
			if(storableId != null){
				return atom.GetStorableByID(storableId);
			}
			else return null;
		}
	}

	public class FloatDriver
	{
		public JSONStorableBool enabled = new JSONStorableBool("enabled", true);
		Func<float> GetRate;
        public JSONStorableFloat currentRate = new JSONStorableFloat("activeness", 0f, 0f, 3f, false, false);
        
		public JSONStorableFloat receiver;
		public JSONStorableFloat maxGloss = new JSONStorableFloat("maxGloss", 8f, 4.7f, 8f, true, true);
		public JSONStorableFloat minGloss = new JSONStorableFloat("minGloss", 4.7f, 0f, 7.5f, true, true);
		public JSONStorableFloat thresholdAtMax;
		public JSONStorableFloat thresholdAtMin;
		public JSONStorableFloat exhaustionEffect = new JSONStorableFloat("Exhaustion effectiveness", .5f, 0f, 5f, false, true);
		public JSONStorableFloat baseChangeRatePos = new JSONStorableFloat("Base change rate (increasing)", 2f, 0f, 10f, false, true);
		public JSONStorableFloat baseChangeRateNeg = new JSONStorableFloat("Base change rate (decreasing)", 2f, 0f, 10f, false, true);
		public float timeActive = 0f;
		public float growthThreshold = 0f;
		public JSONStorableString glossInfo = new JSONStorableString("glossInfo", "");
		public JSONStorableString growthThresholdInfo = new JSONStorableString("growthThresholdInfo", "");
		public JSONStorableString timeActiveInfo = new JSONStorableString("timeActiveInfo", "");

		public FloatDriver(Func<float> GetRate, JSONStorableFloat receiver){
			this.GetRate = GetRate;
			this.receiver = receiver;
			thresholdAtMax = new JSONStorableFloat("Threshold at maxGloss", 1.5f, .4f, 5f, false, true);
			thresholdAtMin = new JSONStorableFloat("Threshold at minGloss", .4f, x => thresholdAtMax.min = x, 0f, 8f, false, true);
		}

        public void SetToDefault(){
            maxGloss.val = maxGloss.defaultVal;
            minGloss.val = minGloss.defaultVal;
            thresholdAtMax.val = thresholdAtMax.defaultVal;
            thresholdAtMin.val = thresholdAtMin.defaultVal;
            exhaustionEffect.val = exhaustionEffect.defaultVal;
        }

		List<float> cache = new List<float>{0f,0f,0f,0f,0f,0f,0f,0f,0f,0f};
		public void Update(){
			if(!SuperController.singleton || !SuperController.singleton.freezeAnimation){
				
                currentRate.val = GetMeanActivity();
				if(currentRate.val > thresholdAtMin.val) timeActive += Time.fixedDeltaTime;
				else{
					if(timeActive > 0f) timeActive -= 3f*Time.fixedDeltaTime;
					else timeActive = 0f;
				}
				float highT;
				if(exhaustionEffect.val == 0f) highT = thresholdAtMax.val;
				else highT = thresholdAtMin.val + (thresholdAtMax.val)/(timeActive*exhaustionEffect.val/100f+(thresholdAtMax.val)/(thresholdAtMax.val-thresholdAtMin.val));
				growthThreshold = thresholdAtMin.val + (highT-thresholdAtMin.val)*(receiver.val - minGloss.val) / (maxGloss.val - minGloss.val);
				float goal;
				float baseChangeRate;
				if(currentRate.val > growthThreshold){
					goal = maxGloss.val;
					baseChangeRate = baseChangeRatePos.val;
				}
				else{
					goal = minGloss.val;
					baseChangeRate = baseChangeRateNeg.val;
				}
				
				receiver.val = Mathf.LerpUnclamped(receiver.val, goal, .01f*baseChangeRate * Time.fixedDeltaTime * Math.Abs(currentRate.val- growthThreshold));

				glossInfo.val = "Current gloss: "+receiver.val.ToString("0.000");
				growthThresholdInfo.val = "Current growth threshold: "+growthThreshold.ToString("0.00");
				timeActiveInfo.val = "Time active (exhaustion): "+timeActive.ToString("0.0");
			}
		}

		public float GetMeanActivity(){
			float sum = 0f;
			float newVal = GetRate();
			for(int i = 0; i < 9; i++){
				sum += cache[i+1];
				cache[i] = cache[i+1];
			}
			cache[9] = newVal;
			sum += newVal;
			return sum  / 10;
		}

        public JSONClass GetJSON(){
            JSONClass jc = new JSONClass();
            if(maxGloss.val != maxGloss.defaultVal) jc[maxGloss.name].AsFloat = maxGloss.val;
            if(minGloss.val != minGloss.defaultVal) jc[minGloss.name].AsFloat = minGloss.val;
            if(thresholdAtMax.val != thresholdAtMax.defaultVal) jc[thresholdAtMax.name].AsFloat = thresholdAtMax.val;
            if(thresholdAtMin.val != thresholdAtMin.defaultVal) jc[thresholdAtMin.name].AsFloat = thresholdAtMin.val;
            if(exhaustionEffect.val != exhaustionEffect.defaultVal) jc[exhaustionEffect.name].AsFloat = exhaustionEffect.val;
			if(baseChangeRatePos.val != baseChangeRatePos.defaultVal) jc[baseChangeRatePos.name].AsFloat = baseChangeRatePos.val;
			if(baseChangeRateNeg.val != baseChangeRateNeg.defaultVal) jc[baseChangeRateNeg.name].AsFloat = baseChangeRateNeg.val;
            return jc;
        }

        public void LoadJSON(JSONClass jc){
            if(jc.HasKey(maxGloss.name)) maxGloss.val = jc[maxGloss.name].AsFloat;
            if(jc.HasKey(minGloss.name)) minGloss.val = jc[minGloss.name].AsFloat;
            if(jc.HasKey(thresholdAtMax.name)) thresholdAtMax.val = jc[thresholdAtMax.name].AsFloat;
            if(jc.HasKey(thresholdAtMin.name)) thresholdAtMin.val = jc[thresholdAtMin.name].AsFloat;
            if(jc.HasKey(exhaustionEffect.name)) exhaustionEffect.val = jc[exhaustionEffect.name].AsFloat;
			if(jc.HasKey(baseChangeRatePos.name)) baseChangeRatePos.val = jc[baseChangeRatePos.name].AsFloat;
			if(jc.HasKey(baseChangeRateNeg.name)) baseChangeRateNeg.val = jc[baseChangeRateNeg.name].AsFloat;
        }
    }

    public static class RigidbodyExt{
		static Vector3 lastVelocity = new Vector3();
		static Vector3 lastAngularVelocity = new Vector3();
		public static Vector3 GetAcceleration(this Rigidbody rb){
			Vector3 acceleration = (rb.velocity - lastVelocity) / Time.fixedDeltaTime;
			lastVelocity = rb.velocity;
			return acceleration;
		}
		public static Vector3 GetAngularAcceleration(this Rigidbody rb){
			Vector3 angularAcceleration = (rb.angularVelocity - lastAngularVelocity) / Time.fixedDeltaTime;
			lastAngularVelocity = rb.angularVelocity;
			return angularAcceleration;
		}
	}
}